/** -*- C++ -*-
    @file predicate/matchers.h
    @author Peter Rockai <me@mornfall.net>
    @author Enrico Zini
*/

#include <ept/predicate/predicate.h>
#include <ept/cache/package.h>
#include <ept/cache/tag.h>
#include <wibble/operators.h>
#include <functional>
#include <cctype>

#ifndef APTFRONT_PREDICATE_MATCHERS_H
#define APTFRONT_PREDICATE_MATCHERS_H

namespace ept {
namespace predicate {
namespace match {

using namespace wibble::operators;

template< typename T, typename Self >
struct Match : predicate::PredicateMixin< T, Self >
{
    bool operator<=( const Self &s ) const {
        return this->self().pattern() <= s.pattern();
    }
};

/**
 * Match the exact value.
 *
 * It can work with any type that allows comparison.
 */
template<typename T>
struct Exact : Match< T, Exact< T > >
{
protected:
	T m_pattern;
public:
    Exact(const T& pattern) : m_pattern(pattern) {}

	bool operator()( const T& item )
	{
        return m_pattern == item;
	}

    T pattern() const { return m_pattern; }
};

template< typename T >
struct LessThan: Match< T, LessThan< T > >
{
    LessThan( const T& pattern ): m_pattern( pattern ) {}

    bool operator()( const T &item ) { return item < m_pattern; }

    T pattern() const { return m_pattern; }

protected:
    T m_pattern;
};

/**
 * Match a substring.
 */
struct Substring : Match< std::string, Substring >
{
protected:
	std::string m_pattern;
public:
    // work around some weird ovelroad resolution conflict with some weird
    // "tolower" which noone knows
    static char mytolower( char c ) { return std::tolower( c ); }

    Substring(const std::string& pattern) {
        std::transform( pattern.begin(), pattern.end(),
                        std::back_inserter( m_pattern ), mytolower );
    }

	bool operator()( const std::string& str )
	{
        std::string in;
        std::transform( str.begin(), str.end(), std::back_inserter( in ), mytolower );
        return in.find( m_pattern, 0 ) != std::string::npos;
	}

    std::string pattern() const { return m_pattern; }

};

/**
 * Match a regular expression.
 *
 * (still unfinished)
 */
struct Ereg : Match<std::string, Ereg>
{
protected:
	std::string m_pattern;
public:
    Ereg(const std::string& pattern) : m_pattern(pattern) {}

	bool operator()( const std::string& str )
	{
        return str.find( m_pattern, 0 ) != std::string::npos;
	}

    std::string pattern() const { return m_pattern; }
};


/**
 * Match tag sets containing the given tags.
 */
template< typename Tag, typename M = Tag >
struct TagMatch : Match< typename Tag::Set, TagMatch< Tag, M > >
{
protected:
	M m_tag;
public:
	TagMatch( const M& tag ) : m_tag(tag) {}

	bool operator()( const typename Tag::Set &tags )
	{
		return m_tag <= tags;
	}

	M pattern() const { return m_tag; }
};

// typedef TagMatch< cache::entity::Tag > Tag;
// typedef TagMatch< cache::entity::Tag::Set > TagSet;

/*
class TagExpr : public Tagcoll::Expression
{
public:
	TagExpr(const std::string expr) : Tagcoll::Expression(expr) {}
	TagExpr(const TagExpr& expr) : Tagcoll::Expression(expr) {}

	typedef Tagcoll::OpSet<cache::entity::Tag> argument_type;
	typedef bool result_type;

    bool operator==( const TagExpr & ) const { return false; } // fixme

    void parseArguments( const ArgumentList & ) {}
    }; */

}

/**
 * Generic predicate for matching.
 *
 * It puts together the machinery of extracting the value from the items using
 * a getter function and of invoking the matching functor on it.
 */
template<typename T, typename Matcher, typename Pattern = typename Matcher::argument_type>
struct AttributeAdaptor : PredicateMixin< T, AttributeAdaptor< T, Matcher, Pattern > >
{
	typedef typename Matcher::argument_type matched_type;

protected:
    typedef matched_type (T::*Getter)( wibble::Fallback< matched_type > ) const;
    Getter m_getter;
    Matcher m_match;
public:
    typedef T argument_type;
    typedef bool result_type;

    AttributeAdaptor( Getter g, const Pattern& pattern )
        : m_getter( g ), m_match( pattern ) {}

    bool operator()( const T &t ) {
        return m_match((t.*m_getter)(typename Matcher::argument_type()));
    }

    bool operator<=( const AttributeAdaptor &a ) const {
        return m_getter == a.m_getter && m_match <= a.m_match;
    }
};

template< typename T >
struct MemberAdaptor : PredicateMixin< T, MemberAdaptor< T > >
{
    typedef bool (T::*Member)( void ) const;

    MemberAdaptor( Member m ) : m_member( m ) {}

    bool operator()( const T &t ) {
        return (t.*m_member)();
    }

    std::string serialize() const {
        return "#error#";
    }

    bool operator<=( const MemberAdaptor &a ) const {
        return m_member == a.m_member;
    }
protected:
    Member m_member;
};

/**
 * Predicate matching tag expressions.
 */
/*
template<typename T>
struct TagexprPredicate : Implementation< T, TagexprMatch< T > >
{
protected:
    typedef Tagcoll::OpSet<cache::entity::Tag> (T::*Getter)( const Tagcoll::OpSet<cache::entity::Tag>& ) const;
    Getter m_getter;
	Tagcoll::TagExpression m_match;
    std::string m_prettyWhat;
public:
    typedef T argument_type;
    typedef bool result_type;

    TagexprPredicate( Getter g, const std::string& pattern, const std::string& w )
        : m_getter( g ), m_match( pattern ), m_prettyWhat( w ) {}

    bool operator()( const T &t ) {
        return m_match((t.*m_getter)(Tagcoll::OpSet<cache::entity::Tag>()));
    }

    std::string prettyPrint() const {
        return m_prettyWhat + ": \"" + m_match.prettyPrint() + "\"";
    }

    std::string serialize() const {
        return std::string( "(" ) + m_prettyWhat + " \"" + m_match.prettyPrint() + "\")";
    }

    void parseArguments( const ArgumentList &a ) {
        m_match.parseArguments( a );
    }
};
*/


}
}

// vim:set ts=4 sw=4:
#endif
